#!/bin/sh

if [ -t 1 ]
then
  BOLD="[1m"
  NORMAL="[0m"
  GOOD="[1;32m"
  BAD="[1;31m"
else
  BOLD=""
  NORMAL=""
  GOOD=""
  BAD=""
fi

cd `dirname $0`/.. || exit 1
echo "entering `pwd`"

die () {
  echo "${BOLD}camical/test/run.sh:${NORMAL} ${BAD}error:${NORMAL} $*" 1>&2
  exit 1
}

[ $# -gt 0 ] && \
die "no command line options expected"

[ -f makefile ] || \
die "can not find '../makefile' (run 'configure.sh' first)"

[ -f ../aiger/aigsim ] || \
die "can not find '../aiger/aigsim' (compile it first)"

make || \
die "'make' failed"

count=0

solvers=""
[ x"`./camical --help|grep cryptominisat`" = x ] || solvers="--cryptominisat"
[ x"`./camical --help|grep ipasir`" = x ] || solvers="$solvers --ipasir"
[ x"`./camical --help|grep riss`" = x ] || solvers="$solvers --riss"

run () {
  for option in \
    "--restore" \
    "--freeze" \
    "--freeze --check" \
    "--hybrid" \
    "--no-incremental" \
    "--no-incremental --no-assume" \
    "--no-restoreall" \
    $solvers
  do

    [ -f test/$1.aig ] || die "can not find 'test/$1.aig'"

    rm -f test/$1.log test/$1.err
    echo "./camical $option test/$1.aig # $2"
    ./camical $option test/$1.aig test/$1.sol 1>test/$1.log 2>test/$1.err
    res=$?

    [ $res = $2 ] || \
    die "expected status '$2' (but got '$res' for 'test/$1.aig')"

    echo "../aiger/aigsim -c test/$1.aig test/$1.sol"
    ../aiger/aigsim -c test/$1.aig test/$1.sol || \
    die "invalid witness 'test/$1.sol' for 'test/$1.aig'"

    count=`expr $count + 1`
  done
}

run true 10
run false 20
run nand 10

run regr0000 10
run regr0001 10
run regr0002 10
run regr0003 10

run cnt1 10
run cnt2 10
run cnt3 10

run cnt1e 10
run cnt2e 10
run cnt3e 10

run cnt1re 10
run cnt2re 10
run cnt3re 10

run dp2 10
run dp3 10
run dp4 10

run miter 20

echo "test results: ${GOOD}$count ok${NORMAL}"
exit 0
