extern "C" {
  #include "aiger/aiger.h"
  #include "aiger/simpaig.h"
}

#include <unistd.h>
#include <sys/types.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <assert.h>
#include <signal.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <cctype>
#include <ctime>

using namespace std;

      
extern "C" {
  typedef struct LatchOrInput LatchOrInput;

  struct LatchOrInput
  {
    unsigned idx;			/* AIGER variable index */
    simpaig *aig;
  };

}

/*
simpaig* simpaig_model;
static simpaigmgr *mgr;

static LatchOrInput *lois;
vector<bool>* latch_cur_vals;
vector<vector<bool>* >* latch_vals;
*/

aiger *aiger_model;
aiger *k_witness_model;
unsigned k;
unsigned num_inputs;
unsigned num_latches;
unsigned num_ands;
unsigned total_number_literals;

unsigned kw_num_inputs;
unsigned kw_num_latches;
unsigned kw_num_ands;
unsigned kw_total_number_literals;

bool is_input(unsigned index){
  unsigned offset = index/2;
  return offset <= num_inputs;
}

bool is_latch(unsigned index) {
  unsigned offset = index/2;
  return offset>num_inputs
      && offset<=(num_inputs + num_latches);
}

bool is_and(unsigned index){
  unsigned offset = index/2;
  return offset>num_inputs + num_latches
      && offset <= total_number_literals;
}

unsigned neg(unsigned l){
  if (l%2 == 1) return l-1;
  return l+1;
}

/* for i-th copy of the circuits, given original index
 * inputs: index + num_inputs * complement * 2
 * latches: index + (k-1)*num_inputs*2 + num_latches * complement * 2
 * ands: index + (k-1)*num_inputs*2 + (k-1)*num_latches*2 + k * 2 + num_ands * complement * 2
 * */
unsigned mapping_old_index_to_k_witness_circuit(unsigned index, unsigned i){
  if(index == 0 || index == 1) return index;
  unsigned complement = (k-1) - i;
  unsigned new_index;
  if(is_input(index)){
    if(i==k-1)
        new_index = index;
    else
        new_index = index + k * num_latches *2 + complement * num_inputs * 2;
  }else if(is_latch(index)){
    new_index = index + complement * num_latches * 2;
  }else if(is_and(index)){
    new_index = index + (k-1)* num_inputs *2 + (k-1)*num_latches*2 + k*2 + num_ands * complement * 2;
  }
  return new_index;
}

bool kw_is_input(unsigned index){
  unsigned offset = index/2;
  return offset <= kw_num_inputs;
}

bool kw_is_latch(unsigned index){
  unsigned offset = index/2;
  return offset>kw_num_inputs
         && offset<=(kw_num_inputs + kw_num_latches);
}

bool kw_is_and(unsigned index){
  unsigned offset = index/2;
  return offset>kw_num_inputs + kw_num_latches
         && offset <= kw_total_number_literals;
}

unsigned map_original(unsigned l){
  if(l==0 || l==1) return l;
  if(kw_is_input(l)){
    return l;
  }else if(kw_is_latch(l)){
    return l + kw_num_inputs * 2;
  }else if(kw_is_and(l)){
    return l + (kw_num_inputs + kw_num_latches) * 2;
  }
}


unsigned map_to_next(unsigned l){
  if(l==0 || l==1) return l;
  if(kw_is_input(l)){
    cout<<"l:"<<l<<endl;
    return l + kw_num_inputs * 2;
  }else if(kw_is_latch(l)){
    unsigned idx = l/2;
    aiger_symbol* s = aiger_is_latch(k_witness_model, l);
    cout<<"l:"<<l<<" "<<"s->lit: "<<s->lit<<endl;
    unsigned new_next = map_original(s->next);
    return l%2? neg(new_next) : new_next;
  }else if(kw_is_and(l)){
    cout<<"land :"<<l<<endl;
    return l + kw_total_number_literals * 2;
  }
}

int 
main (int argc, char *argv[])
{
  /* load aiger file */
  std::string k_val("-k");
  /* output file */
  std::string output_file("-o");

  k = -1;
  const char* output_name;
  const char* input_name;

  for (int i = 1; i < argc - 1; ++i) {
    if(k_val == argv[i]){
      string s = argv[++i];
      k = atoi(s.c_str());
    }else if (output_file == argv[i]){
      output_name = argv[++i];
    }
  }
  assert(k>1);
  assert(argc>=3);
  input_name = argv[argc-1];
  cout<<"Reading file: "<<input_name<<endl;

  aiger_model = aiger_init(); 

  aiger_open_and_read_from_file(aiger_model, input_name);

  num_inputs = aiger_model->num_inputs;
  num_latches = aiger_model->num_latches;
  num_ands = aiger_model->num_ands;
  total_number_literals = num_inputs + num_latches + num_ands;
  assert(num_latches > 0);

  cout<<"inputs="<<num_inputs<<", latches="<<num_latches<<", and_gates="<<num_ands<<endl;

  aiger_and *ands = aiger_model->ands;
  aiger_symbol *latches = aiger_model->latches;
  aiger_symbol *inputs = aiger_model->inputs;

  k_witness_model = aiger_init();

  /* indices in the k-witness aiger model */
  unsigned L_k_1_begin;
  unsigned L_0_begin;
  unsigned B_k_1_begin;
  unsigned and_gate_begin;
  //unsigned uniqueness_begin;
  //unsigned uniqueness;
  unsigned R_i_size;
  unsigned R_or_1_to_k_1;
  unsigned Rp_0_begin;
  unsigned original_output;
  unsigned original_p;
  unsigned output;
  vector<unsigned>* Rs = new vector<unsigned>();
  vector<unsigned>* hs = new vector<unsigned>(); /* size of k-1, from 1..k-1, F^i = F(L^{i-1})*/
  vector<unsigned>* ps = new vector<unsigned>(); /* the properties */

  unsigned num_comparisons = k*(k-1)/2;

  L_k_1_begin = num_inputs * 2 + 2;
  L_0_begin = L_k_1_begin + (k-1) * num_latches * 2;
  B_k_1_begin = (num_inputs + num_latches) * k * 2 + 2;
  and_gate_begin = (k * (num_inputs + num_latches) + k ) * 2 + 2;

  //uniqueness_begin = and_gate_begin + num_ands * k * 2;
  //uniqueness = uniqueness_begin + (4 * num_latches -1) * k * (k-1) - 2 + (num_comparisons-1)*2;
  R_i_size = (3*num_latches + num_latches - 1) *2;
  aiger_symbol* o_s = aiger_model->outputs;
  original_output = o_s->lit;
  original_p = neg(original_output);
  /* only one output in the original circuit */
  assert(aiger_model->num_outputs == 1);

  cout<<"L_k_1_begin="<<L_k_1_begin<<endl;
  cout<<"L_0_begin="<<L_0_begin<<endl;
  cout<<"B_k_1_begin="<<B_k_1_begin<<endl;
  cout<<"original property="<<original_p<<endl;


    /* adding original set of inputs */
    for(unsigned i=0; i<num_inputs; i++){
        aiger_add_input(k_witness_model, (i+1)*2, "");
    }

   /* adding k copies of latches */
    /* copying L_{k-1} from original circuit */
    for(unsigned i=0; i<num_latches; i++){
      aiger_symbol* current_latch = latches + i;
      unsigned latch_index = mapping_old_index_to_k_witness_circuit(current_latch->lit, k-1);
      unsigned next  = mapping_old_index_to_k_witness_circuit(current_latch->next, k-1);
      unsigned reset  = mapping_old_index_to_k_witness_circuit(current_latch->reset, k-1);
      aiger_add_latch(k_witness_model, latch_index, next, "");
      aiger_add_reset(k_witness_model, latch_index, reset);
    }

    cout<<"added L^{k-1}"<<endl;

    /* adding the rest k-1 copies of latches */
      for(unsigned i=k-2; i!=-1; i--){
        for(unsigned j=0; j<num_latches; j++){
          aiger_symbol* current_latch = latches + j;

          unsigned latch_index = mapping_old_index_to_k_witness_circuit(current_latch->lit, i);
          if(i==0 && j==0){
            L_0_begin = current_latch->lit;
          }
          unsigned next = mapping_old_index_to_k_witness_circuit(current_latch->lit, i+1);
          aiger_add_latch(k_witness_model, latch_index, next, "");
        }
      }

      cout<<"added k copies of latches"<<endl;

  /* adding k-1 copies of inputs, as latches */
  for(unsigned i=k-2; i!=-1; i--){

    for(unsigned j=0; j<num_inputs; j++){
      aiger_symbol* current_var = inputs + j;
      unsigned input_index = mapping_old_index_to_k_witness_circuit(current_var->lit, i);
      unsigned next = mapping_old_index_to_k_witness_circuit(current_var->lit, i+1);
      aiger_add_latch(k_witness_model, input_index, next, "");
      aiger_add_reset(k_witness_model, input_index, input_index);
    }
  }

  cout<<"added k copies of inputs"<<endl;



  /* adding B */
  //aiger_add_latch(k_witness_model, B_k_1_begin, uniqueness + 2, "");
  aiger_add_latch(k_witness_model, B_k_1_begin, B_k_1_begin, "");
  for(unsigned i=1; i<k; i++){
    unsigned index = B_k_1_begin + i*2;
    //unsigned next = uniqueness + 2*(i+1);
    unsigned next = B_k_1_begin + (i-1)*2;
    aiger_add_latch(k_witness_model, index, next, "");
  }

  /* copying the original circuit k times */
  unsigned current_index;
  for(unsigned i=k-1; i!=-1; i--){
    //cout<<"copying the "<<complement<<"th circuit,"<<endl;
    for(unsigned j=0; j<num_ands; j++){
      aiger_and* original_and = ands + j;
      unsigned old_index;
      unsigned old_rhs0;
      unsigned old_rhs1;

      old_index = original_and->lhs;
      old_rhs0 = original_and->rhs0;
      old_rhs1 = original_and->rhs1;
      /* new and-gate */
      unsigned and_index;
      unsigned rhs0;
      unsigned rhs1;

      and_index = mapping_old_index_to_k_witness_circuit(old_index, i);
      rhs0 = mapping_old_index_to_k_witness_circuit(old_rhs0, i);
      rhs1 = mapping_old_index_to_k_witness_circuit(old_rhs1, i);

      aiger_add_and(k_witness_model, and_index, rhs0, rhs1);
      //cout<<"and "<<and_index<<" "<<rhs0<<" "<<rhs1<<endl;
      current_index = and_index;
    }
  }

  cout<<"added k copies of original circuits"<<endl;
  current_index += 2;
  /* construct R(L^0)...R(L^{k-1}) */
  cout<<"constructing R0... Rk-1"<<endl;
  for(unsigned i=0; i<k; i++){
    vector<unsigned>* latch_reset_equivs = new vector<unsigned>();

    unsigned complement = k-1-i;
    /* L_i_begin: latch literal */
    unsigned L_i_begin = L_k_1_begin + (complement) * num_latches * 2;
    for(unsigned j=0; j<num_latches; j++){
      unsigned latch_literal = L_i_begin + j * 2;
      aiger_symbol* current_latch = latches + j;
      unsigned reset = mapping_old_index_to_k_witness_circuit(current_latch->reset, i);

      aiger_add_and(k_witness_model, current_index, latch_literal, neg(reset)); /* l ^ 1 */
      current_index += 2;
      aiger_add_and(k_witness_model, current_index, latch_literal + 1, reset); /* -l ^ 0 */
      current_index += 2;
      aiger_add_and(k_witness_model, current_index,
              current_index - 3, current_index - 1);
      latch_reset_equivs->push_back(current_index);
      current_index += 2;
    }
    unsigned prev = latch_reset_equivs->at(0);
    if(num_latches == 1){
      Rs->push_back(prev);
    }else{

      for(unsigned i=1; i<latch_reset_equivs->size(); i++){
        unsigned x = latch_reset_equivs->at(i);
        aiger_add_and(k_witness_model, current_index, prev, x);
        if(i==latch_reset_equivs->size()-1){
          Rs->push_back(current_index);
        }
        prev = current_index;
        current_index += 2;
      }
    }
  }

  cout<<"added R(L^i)s"<<endl;

  /* Resetting the latches */
  /* Resetting R_0 */
  /* R_or_1_to_k_1: V_{1,k-1} R^i, condition for Rp_0 (R'_0)*/
  vector<int>* R_or_i_to_k_1s = new vector<int>;
  assert(Rs->size()>1);
  unsigned prev = Rs->at(1);
  /* the model is 2-inductive */
  if(Rs->size() == 2){
    R_or_1_to_k_1 = prev;
    current_index = prev + 2;
  }else {
    /* R_or_k_1_to_1 */
    prev = Rs->at(Rs->size()-1) + 1;
    for(unsigned i=k-2; i!=0; i--){
      unsigned R_i = Rs->at(i);
      aiger_add_and(k_witness_model, current_index, prev, R_i + 1);
      prev = current_index;
      current_index += 2;
      R_or_i_to_k_1s->insert(R_or_i_to_k_1s->begin(),current_index-1);
    }

    R_or_1_to_k_1 = current_index-1;

  }

  cout<<"R'0 reset"<<endl;

  /* R'(L^0): (R_or_1_to_k_1 ^ l) V (R_or_1_to_k_1 ^ r_l) */
  for(unsigned i=0; i<num_latches; i++){
    aiger_symbol* current_latch = latches + i;
    unsigned latch_index = L_0_begin + i*2;
    aiger_add_and(k_witness_model, current_index, R_or_1_to_k_1, latch_index);
    current_index += 2;
    aiger_add_and(k_witness_model, current_index, neg(R_or_1_to_k_1), current_latch->reset);
    current_index += 2;
    aiger_add_and(k_witness_model, current_index, current_index-3, current_index-1);
    //aiger_add_reset(k_witness_model, latch_index, current_index+1);
    // aiger_add_reset(k_witness_model, latch_index, latch_index);
    current_index += 2;
  }

  /* constructing f(L^1) ... f(L^{k-1}) */
  /* and resettinng R'(L^1)...R'(L^{k-2}), only for k>2 */
  for(unsigned i=1; i<k-1; i++){
    unsigned complement = (k-1) - i;

    unsigned R_or_i_to_k_1;
    if(i!=1){
    unsigned prev = Rs->at(i) + 1;
    for(unsigned j=i+1; j<k; j++){
      unsigned current_R = Rs->at(j);
      aiger_add_and(k_witness_model, current_index, prev, current_R+1);
      prev = current_index;
      current_index += 2;
    }
    //R_or_i_to_k_1 = current_index - 1;}
    R_or_i_to_k_1 = R_or_i_to_k_1s->at(i-1);}
    else{
      R_or_i_to_k_1 = R_or_1_to_k_1;
    }

    /* R'(L^{i-1}) */
    vector<unsigned>* rp = new vector<unsigned>;
    for(unsigned j=0; j<num_latches; j++){
      aiger_symbol* current_latch = latches + j;
      unsigned original_lit = current_latch->lit;
      unsigned L_i_1_lit = mapping_old_index_to_k_witness_circuit(original_lit, i-1);

      aiger_symbol* L_i_1_latch = aiger_is_latch(k_witness_model, L_i_1_lit);

      //assert(L_i_1_latch->lit == L_i_1_lit);
      unsigned L_i_1_reset = L_i_1_latch->reset;
      aiger_add_and(k_witness_model, current_index, L_i_1_lit, neg(L_i_1_reset));
      current_index += 2;
      aiger_add_and(k_witness_model, current_index, L_i_1_lit+1, L_i_1_reset);
      current_index += 2;
      aiger_add_and(k_witness_model, current_index, current_index-3, current_index-1);
      rp->push_back(current_index);
      current_index += 2;
    }

    prev = rp->at(0);
    for(unsigned j=1; j<num_latches; j++){
      aiger_add_and(k_witness_model, current_index, prev, rp->at(j));
      prev = current_index;
      current_index += 2;
    }

    unsigned rp_i_1 = current_index-2;

    for(unsigned j=0; j<num_latches; j++){

      aiger_symbol* current_latch = latches + j;
      unsigned original_next = current_latch->next;
      unsigned new_next = mapping_old_index_to_k_witness_circuit(original_next, i-1); /* f^{i-1} */

      aiger_add_and(k_witness_model, current_index, new_next, new_next);
      current_index += 2;

      unsigned L_i_latch_lit = mapping_old_index_to_k_witness_circuit(current_latch->lit, i);

      aiger_add_and(k_witness_model, current_index, R_or_i_to_k_1 - 1, current_index - 2); /* or_R ^ f */
      current_index += 2;
      aiger_add_and(k_witness_model, current_index, R_or_i_to_k_1, L_i_latch_lit); /* or_R ^ l */
      current_index += 2;
      aiger_add_and(k_witness_model, current_index, current_index - 3, current_index - 1); /* or_R ^ l */
      //aiger_add_reset(k_witness_model, L_i_latch_lit, current_index+1);
      aiger_add_reset(k_witness_model, L_i_latch_lit, L_i_latch_lit);

      current_index += 2;
    }
  }

  /* Reset R^{k-1} */
  unsigned R_k_1 = Rs->at(k-1);
  vector<unsigned>* rp = new vector<unsigned>;
  for(unsigned j=0; j<num_latches; j++){
    aiger_symbol* current_latch = latches + j;
    unsigned original_lit = current_latch->lit;
    unsigned L_i_1_lit = mapping_old_index_to_k_witness_circuit(original_lit, k-2);
    aiger_symbol* L_i_1_latch = aiger_is_latch(k_witness_model, L_i_1_lit);
    //assert(L_i_1_latch->lit == L_i_1_lit);
    unsigned L_i_1_reset = L_i_1_latch->reset;
    aiger_add_and(k_witness_model, current_index, L_i_1_lit, neg(L_i_1_reset));
    current_index += 2;
    aiger_add_and(k_witness_model, current_index, L_i_1_lit+1, L_i_1_reset);
    current_index += 2;
    aiger_add_and(k_witness_model, current_index, current_index-3, current_index-1);
    rp->push_back(current_index);
    current_index += 2;
  }

  prev = rp->at(0);
  for(unsigned j=1; j<num_latches; j++){
    aiger_add_and(k_witness_model, current_index, prev, rp->at(j));
    prev = current_index;
    current_index += 2;
  }

  unsigned rp_i_1 = current_index-2;
  for(unsigned i=0; i<num_latches; i++){
    aiger_symbol* current_latch = latches + i;
    unsigned original_next = current_latch->next;
    unsigned new_next = mapping_old_index_to_k_witness_circuit(original_next, k-2); /* f^{i-1} */
    aiger_add_and(k_witness_model, current_index, new_next, new_next);
    unsigned f = current_index;
    current_index += 2;

    unsigned l = mapping_old_index_to_k_witness_circuit(current_latch->lit, k-1);
    aiger_add_and(k_witness_model, current_index, R_k_1, l);
    current_index += 2;
    aiger_add_and(k_witness_model, current_index, neg(R_k_1), f);
    current_index += 2;
    aiger_add_and(k_witness_model, current_index, current_index-3, current_index-1);

    //aiger_add_reset(k_witness_model, l, current_index+1);
    // aiger_add_reset(k_witness_model, l, current_latch->reset);
    current_index += 2;
  }

  /* Reset B */
  /* Reset b^{k-1} */
  aiger_add_reset(k_witness_model, B_k_1_begin, 1);
  /* Reset b^i for i in 1...k-2 */
  cout<<"initialising Bi"<<endl;
  for(unsigned i=1 ; i<k-1; i++){
    unsigned complement = (k-1) - i;
    unsigned b_i = B_k_1_begin + complement * 2;
    unsigned b_i_1 = B_k_1_begin + (complement+1) * 2;
    unsigned R_i = Rs->at(i);

    unsigned prev1 = neg(Rs->at(i+1));
    
    //for(unsigned j=i+2; j<k; j++){
    //  aiger_add_and(k_witness_model, current_index, prev1, neg(Rs->at(j)));
    //  prev1 = current_index;
    //  current_index += 2;
    //}
    //unsigned and_not_R_j = i==k-2? prev1: current_index-2;
    unsigned and_not_R_j = i==k-2? prev1: neg(R_or_i_to_k_1s->at(i));
    
    aiger_add_and(k_witness_model, current_index, R_i, and_not_R_j);

    current_index += 2;
    aiger_add_and(k_witness_model, current_index, b_i_1+1, current_index-1);
    //aiger_add_reset(k_witness_model, b_i, current_index+1);
    aiger_add_reset(k_witness_model, b_i, 0);
    current_index += 2;
  }
  /* Reset b^0 */
  //prev = Rs->at(1);
  //prev = neg(prev);
  //for(unsigned i=2; i<k; i++){
  //  unsigned R_i = Rs->at(i);
  //  aiger_add_and(k_witness_model, current_index, prev, neg(R_i));
  //  prev = current_index;
  //  current_index += 2;
  //}
  //unsigned or_R = current_index-1;
  unsigned or_R = R_or_1_to_k_1;
  if(k==2){
    or_R = Rs->at(1);
  }
  aiger_add_and(k_witness_model, current_index, or_R, 0);
  current_index += 2;
  aiger_add_and(k_witness_model, current_index, neg(or_R), 1);
  current_index += 2;
  aiger_add_and(k_witness_model, current_index, current_index-1, current_index-3);
  current_index += 2;
  //aiger_add_reset(k_witness_model, B_k_1_begin + (k-1)*2, neg(or_R));
  aiger_add_reset(k_witness_model, B_k_1_begin+(k-1)*2, 0);
  /* Construct the property */
  /* p3  */
  /* p3: d^i->R(L^{i+1}) for i in 0..k-2 */
  /*
  vector<unsigned>* sis = new vector<unsigned>();
  vector<unsigned>* vis = new vector<unsigned>();

  for(unsigned i=0; i<k-1; i++){
      unsigned complement = (k-1)-i;
      unsigned b_j = B_k_1_begin + complement * 2;
      if(i==0){
        sis->push_back(b_j+1);
      }else{
        aiger_add_and(k_witness_model, current_index, b_j+1, sis->at(i-1));
        sis->push_back(current_index);
        current_index+=2;
      }

  }

  for(unsigned i=0; i<k-1; i++){
      unsigned b_j = B_k_1_begin + i * 2;
      if(i==0){
        vis->push_back(b_j);
      }else{
        aiger_add_and(k_witness_model, current_index, b_j, vis->at(i-1));
        vis->push_back(current_index);
        current_index+=2;
      }

  }
  */

  vector<unsigned>* p3s = new vector<unsigned>();
  for(unsigned i=0; i<k-1; i++){
    unsigned prev;
    //prev = sis->at(i);
    unsigned complement = (k-1)-(i+1);
    //aiger_add_and(k_witness_model, current_index, prev, vis->at(complement));
    unsigned bi = B_k_1_begin + complement *2;
    unsigned bi1 = bi + 2;
    aiger_add_and(k_witness_model, current_index, bi, neg(bi1));
    current_index+=2;
    /*
    for(unsigned j=i+1; j<k; j++){
      unsigned complement = (k-1)-j;
      unsigned b_j = B_k_1_begin + complement * 2;
      aiger_add_and(k_witness_model, current_index, prev, b_j);
      prev = current_index;
      current_index += 2;
    }
    */

    unsigned R_i_p_1 = Rs->at(i+1);
    unsigned R_i_p_1_ = R_i_p_1%2? R_i_p_1-1: R_i_p_1+1;
    aiger_add_and(k_witness_model, current_index, current_index-2, R_i_p_1_);
    p3s->push_back(current_index+1);
    current_index += 2;
  }

  /* p0:  b^i -> b^{i+1} for i in 0...k-2 */
  vector<unsigned>* b_i_i_1 = new vector<unsigned>();
  if(k==2){
    unsigned b_0 = B_k_1_begin + 2;
    aiger_add_and(k_witness_model, current_index, b_0, neg(B_k_1_begin));
    current_index += 2;
    ps->push_back(current_index-1);
  }else {
    for (unsigned i = 0; i < k - 1; i++) {
      unsigned complement = (k - 1) - i;
      unsigned b_i = B_k_1_begin + complement * 2;
      /* b^i ^ !b^{i+1} */
      unsigned bis = b_i-1;
      /*
      if(i+1<k-1){
        unsigned prev;
        for(unsigned j=i+1; j<k;j++){
          unsigned c = (k-1)-j;
          unsigned b_j = B_k_1_begin + c * 2;

          if(j==i+1){
            prev = b_j;
          }else{
            aiger_add_and(k_witness_model, current_index, prev, b_j);
            prev = current_index;
            current_index += 2;
          }

        }
        bis = current_index - 1;
      }
      */
      //or simply, bis=b_i-1;
      aiger_add_and(k_witness_model, current_index, b_i, bis);
      b_i_i_1->push_back(current_index + 1);
      current_index += 2;
    }
    /* constructing p0 */
    assert(b_i_i_1->size() == k - 1);
    prev = b_i_i_1->at(0);
    for (unsigned i = 1; i < k - 1; i++) {
      unsigned index = b_i_i_1->at(i);
      aiger_add_and(k_witness_model, current_index, prev, index);
      prev = current_index;
      current_index += 2;
    }
    ps->push_back(current_index-2);
  }

  /* p1: b^i -> h^i, for i in 1..k-1 */
  vector<unsigned>* p1s = new vector<unsigned>();
  for(unsigned i=0; i<k-1; i++){
    unsigned complement = (k-1) - i;
    unsigned b_i = B_k_1_begin + complement * 2;

    vector<unsigned>* h_ls = new vector<unsigned>();
    for(unsigned j=0; j<num_latches; j++){
      aiger_symbol* original_latch = latches + j;
      unsigned l = mapping_old_index_to_k_witness_circuit(original_latch->lit, i+1);

      unsigned original_next = original_latch->next;
      unsigned new_next_i = mapping_old_index_to_k_witness_circuit(original_next, i);
      /* l<->new_next_i */
      /* l^!new_next_i */
      aiger_add_and(k_witness_model, current_index, l, neg(new_next_i));
      current_index += 2;
      /* !l ^ new_next_i */
      aiger_add_and(k_witness_model, current_index, new_next_i, l+1);
      current_index +=2;
      aiger_add_and(k_witness_model, current_index, current_index-3, current_index-1);
      h_ls->push_back(current_index);
      current_index +=2;
    }
    prev = h_ls->at(0);
    for(unsigned i=1; i<num_latches; i++){
      unsigned index = h_ls->at(i);
      aiger_add_and(k_witness_model, current_index, prev, index);
      prev = current_index;
      current_index += 2;
    }
    /* b^i ^ !h^i*/
    aiger_add_and(k_witness_model, current_index, b_i, current_index-1);
    p1s->push_back(current_index+1);
    current_index += 2;
  }

  /* p1 */
  prev = p1s->at(0);
  if(p1s->size() == 1){
    ps->push_back(prev);
  }else {
    for (unsigned i = 1; i < k - 1; i++) {
      unsigned index = p1s->at(i);
      aiger_add_and(k_witness_model, current_index, prev, index);
      prev = current_index;
      current_index += 2;
    }
    ps->push_back(current_index - 2);
  }

  cout<<"p1 set"<<endl;
  /* p2: b^i -> p^i */
  vector<unsigned>* p2s = new vector<unsigned>();
  for(unsigned i=0; i<k; i++){
    unsigned complement = (k-1) - i;
    unsigned b_i = B_k_1_begin + complement * 2;
    unsigned p_i = mapping_old_index_to_k_witness_circuit(original_p, i);
    aiger_add_and(k_witness_model, current_index, b_i, neg(p_i));
    p2s->push_back(current_index+1);
    current_index += 2;
  }
  /* p2 */
  prev = p2s->at(0);
  for(unsigned i=1; i<k; i++){
    unsigned index = p2s->at(i);
    aiger_add_and(k_witness_model, current_index, prev, index);
    prev = current_index;
    current_index += 2;
  }
  ps->push_back(current_index-2);
  cout<<"p2 set"<<endl;

  
  /* p3 */
  prev = p3s->at(0);
  if(p3s->size() == 1){
    ps->push_back(prev);
  }else {
    for (unsigned i = 1; i < k - 1; i++) {
      unsigned index = p3s->at(i);
      aiger_add_and(k_witness_model, current_index, prev, index);
      prev = current_index;
      current_index += 2;
    }
    ps->push_back(current_index - 2);
  }

  /* p' = p1...p3 */
  prev = ps->at(0);
  for(unsigned i=1; i<ps->size(); i++){
    unsigned index = ps->at(i);
    aiger_add_and(k_witness_model, current_index, prev, index);
    prev = current_index;
    current_index += 2;
  }

  /* p4 */
  aiger_add_and(k_witness_model, current_index, prev, B_k_1_begin);

  unsigned new_property = current_index;
  output = current_index + 1;


  aiger_add_output(k_witness_model, output, "");
  // aiger_reencode(k_witness_model);
  if(output_name){
    aiger_open_and_write_to_file(k_witness_model, output_name);
  }else{
    aiger_write_to_file (k_witness_model, aiger_ascii_mode, stdout);
  }


 
  return 0;
}

