#!/usr/bin/env python3

import re
import sys
import os
import signal
from enum import Enum
from subprocess import Popen, PIPE
from pyparsing import nestedExpr
import pyparsing as pp
import re
import time
import uuid
import logging

modelFile = ""
k=0
kwFile = ""
tempdirectory = "temp"

row_string = ""

def generate_kw_model(k, model):
    global tempdirectory, row_string
    print("generating k-witness model......")
    exe = sys.path[0] + "/aigcertify_kind "
    kflag = "-k " + str(k) + " "
    f = " " + sys.path[0] + "/" + model.name + " "
    outfile = tempdirectory + "/kw-" + str(uuid.uuid4()) + ".aag"

    cmd = exe + kflag + "-o " + outfile + f
    print("cmd: ", cmd)
    time_start = time.time()

    signal.signal(signal.SIGPIPE, signal.SIG_IGN)

    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    time_taken = (time.time() - time_start)
    output = output.splitlines()
       
    outfile1 = open(outfile, "r")
    h = outfile1.readline()
    print("header:", h)
    header = h.split(" ")
    fsize_ands = header[-1]
    fsize_maxvar = format(int(header[1])/1000, '.2f')

    outfile1.close()
    print("kw-model generation time=", time_taken)
    print("maxvar=", fsize_maxvar)
    print("andgates=", fsize_ands)

    logger = logging.getLogger("kw-model-gen")
    logger.info("time="+str(time_taken))
    logger.info("maxvar="+fsize_maxvar)
    logger.info("ands="+fsize_ands)
       
    row_string+=fsize_maxvar +"', '"
    row_string+=format(time_taken,'.2f') + "', '"

    return outfile


def check_comb_sim(k, model, kwmodel):
    global tempdirectory
    exe = sys.path[0] + "/mapping_check/mapping_check "
    kflag = "-k " + str(k) + " "
    mf = " " + sys.path[0] + "/" + model.name + " "
    kmf = " " + sys.path[0] + "/" + kwmodel + " "

    outfile1 = tempdirectory + "/simf-" + str(uuid.uuid4()) + ".aag"
    outfile2 = tempdirectory + "/simprop-" + str(uuid.uuid4()) + ".aag"
    outfile3 = tempdirectory + "/simreset-" + str(uuid.uuid4()) + ".aag"

    cmd = exe  + "-o " + outfile1 + " " + outfile2 +" "+ outfile3 +  mf + kmf
    print("cmd: ", cmd)
    time1 = time.time() 
    signal.signal(signal.SIGPIPE, signal.SIG_IGN);

    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    time_taken = time.time() - time1
    output = output.splitlines()


    outfile11 = open(outfile1, "r")
    header = outfile11.readline().split(" ")
    fsize_ands1 = header[-1]
    fsize_maxvar1 = format(int(header[1])/1000, '.2f')

    outfile11.close()
        
    outfile22 = open(outfile2, "r")
    header = outfile22.readline().split(" ")
    fsize_ands2 = header[-1]
    fsize_maxvar2 = format(int(header[1])/1000, '.2f')
    outfile22.close()

    outfile33 = open(outfile3, "r")
    header = outfile33.readline().split(" ")
    fsize_ands3 = header[-1]
    fsize_maxvar3 = format(int(header[1])/1000, '.2f')
    outfile33.close()
    print("c-simulation check time=", time_taken)
    print("maxvar1=", fsize_maxvar1)
    print("andgates1=", fsize_ands1)
    print("maxvar2=", fsize_maxvar2)
    print("andgates2=", fsize_ands2)
    print("maxvar3=", fsize_maxvar3)
    print("andgates3=", fsize_ands3)

    logger1 = logging.getLogger("c-simulation-check1")
    logger1.info("time="+str(time_taken))
    logger1.info("maxvar="+fsize_maxvar1)
    logger1.info("ands="+fsize_ands1)

    logger2 =  logging.getLogger("c-simulation-check2")
    logger2.info("maxvar="+fsize_maxvar2)
    logger2.info("ands="+fsize_ands2)
        
    logger3 =  logging.getLogger("c-simulation-check3")
    logger3.info("maxvar="+fsize_maxvar3)
    logger3.info("ands="+fsize_ands3)

    return time_taken,outfile1, outfile2, outfile3, fsize_maxvar1, fsize_maxvar2, fsize_maxvar3



def generate_proof(file):
    global tempdirectory

    print("generating three conditions for the invariant...")
    exe = sys.path[0] + "/aigcertify/aigcertify "

    f1 = " " + sys.path[0] + "/" + file
    f2 = " " + sys.path[0] + "/" + file + " "

    out1 = tempdirectory + "/con1-" + str(uuid.uuid4()) + ".aag"
    out2 = tempdirectory + "/con2-" + str(uuid.uuid4()) + ".aag"
    out3 = tempdirectory + "/con3-" + str(uuid.uuid4()) + ".aag"
    cmd = exe + "-o " + out1 + " " + out2 + " " + out3 + f1 + f2
    print("cmd: ", cmd)
        
    time_start = time.time()
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    time_taken = time.time() - time_start
    output = output.splitlines()

    outfile1 = open(out1, "r")
    header = outfile1.readline().split(" ")
    fsize_ands1 = header[-1]
    fsize_maxvar1 = format(int(header[1])/1000, '.2f')
    outfile1.close()

    outfile2 = open(out2, "r")
    header = outfile2.readline().split(" ")
    fsize_ands2 = header[-1]
    fsize_maxvar2 = format(int(header[1])/1000, '.2f')
    outfile2.close()

    outfile3 = open(out3, "r")
    header = outfile3.readline().split()
    fsize_ands3 = header[-1]
    fsize_maxvar3 = format(int(header[1])/1000, '.2f')
    outfile3.close()

    print("certification  generation time=", time_taken)
    print("maxvar1=", fsize_maxvar1)
    print("maxvar2=", fsize_maxvar2)
    print("maxvar3=", fsize_maxvar3)
    print("andgates1=", fsize_ands1)
    print("andgates2=", fsize_ands2)
    print("andgates3=", fsize_ands3)
    logf = open(sys.path[0] + "/temp/logfile.txt", "w")
    logf.write("3cert-gen="+str(time_taken))
    logf.write("cert1-maxvar="+fsize_maxvar1)
    logf.write("cert2-maxvar="+fsize_maxvar2)
    logf.write("cert3-maxvar="+fsize_maxvar3)
    logf.write("cert1-ands="+fsize_ands1)
    logf.write("cert2-ands="+fsize_ands2)
    logf.write("cert3-ands="+fsize_ands3)
    logf.close()


    logger1 = logging.getLogger("aigcertify1")
    logger1.info("time="+str(time_taken))
    logger1.info("maxvar="+fsize_maxvar1)
    logger1.info("ands="+fsize_ands1)

    logger2 = logging.getLogger("aigcertify2")
    logger2.info("maxvar="+fsize_maxvar2)
    logger2.info("ands="+fsize_ands2)

    logger3 = logging.getLogger("aigcertify3")
    logger3.info("maxvar="+fsize_maxvar3)
    logger3.info("ands="+fsize_ands3)

    return time_taken, out1, out2, out3, fsize_maxvar1, fsize_maxvar2, fsize_maxvar3


def aiger2cnf(file):
     global tempdirectory

     print("aiger to cnf...")
     exe = sys.path[0] + "/aiger/aigtocnf "

     f = " " + sys.path[0] + "/" + file + " "
     out = tempdirectory + "/cnf-" + str(uuid.uuid4())

     cmd = exe + f + out
     print("cmd: ", cmd)

     signal.signal(signal.SIGPIPE, signal.SIG_IGN);

     time_start = time.time()
     p = Popen(cmd, shell=True, stdout=PIPE)
     output = p.communicate()[0]
     time_taken = time.time()-time_start
     output = output.splitlines()

     return time_taken,out

def verify_qbf(file, idx):
    global tempdirector, row_string
    print("QAIGER to QCIR...")
    exe = sys.path[0]+"/qaiger2qcir "
    f = " " + sys.path[0] + "/" + file + " "
    out = tempdirectory + "/qcir-" + str(uuid.uuid4())
      
    cmd = exe + f +" >> "+ out
    print("cmd: ", cmd)

    signal.signal(signal.SIGPIPE, signal.SIG_IGN);

    time_start = time.time()
    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    time_taken = time.time()-time_start
    output = output.splitlines()
    
    exe = sys.path[0]+"/quabs "
    f = " " + sys.path[0] + "/" + out + " "
    
    cmd = exe + f     
    print("cmd: ", cmd)

    signal.signal(signal.SIGPIPE, signal.SIG_IGN);

    time_start2 = time.time()
    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    time_taken2 = time.time()-time_start2
    time_taken = time_taken2 + time_taken
    output = output.splitlines()

    res = output[0].decode('utf8')
    print(res)

    logger1 = logging.getLogger("quabs" + str(idx))
    logger1.info("time="+str(time_taken))
    row_string += ", '" + format(time_taken, '.2f') +"', '" + format(time_taken2, '.2f') +"'"
    return res

def picosat_sat(file, idx):

    exe = sys.path[0] + "/picosat/picosat "

    f = " " + sys.path[0] + "/" + file + " "

    cmd = exe + f
    print("cmd: ", cmd)
    
    time_start = time.time()

    signal.signal(signal.SIGPIPE, signal.SIG_IGN);

    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]

    time_taken = (time.time() - time_start)

    output = output.splitlines()

    res = output[0].decode('utf8')
    print(res)
    print("picosat sovling time: ", time_taken)

    logger1 = logging.getLogger("picosat" + str(idx))
    logger1.info("time="+str(time_taken))

    return res

def kissat_sat(file,idx):
    global row_string
    exe = sys.path[0] + "/../kissat/build/kissat -q "

    f = " " + sys.path[0] + "/" + file + " "

    cmd = exe + f
    print("cmd: ", cmd)

    time_start = time.time()

    signal.signal(signal.SIGPIPE, signal.SIG_IGN);

    p = Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]
    time_taken = time.time() - time_start
    output = output.splitlines()
    res = output[0].decode('utf8')

    print(res)
    print("kissat solving time: ", time_taken)
    
    logger1 = logging.getLogger("kissat" + str(idx))
    logger1.info("time="+str(time_taken))
    row_string += ", '" + format(time_taken, '.2f') +"'"
    return res

def sat(file, idx):
    if use_kissat==1:
        return kissat_sat(file, idx)
    else:
        return picosat_sat(file, idx)

def mcaiger(f, model):
    global row_string
    print("calling Mcaiger...")
    exe = sys.path[0] + "/mcaiger/mcaiger -n "
    f = " " + sys.path[0] + "/" + model.name + " "

    cmd = exe + f
    print("cmd: ", cmd)
    
    time1 = time.time()
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    p=Popen(cmd, shell=True, stdout=PIPE)
    output = p.communicate()[0]

    time_taken = time.time()-time1
    output = output.splitlines()
    k = output[-1].decode('utf8').split("=")[-1]
   
    relat_name = model.name.split("/")[-1].split(".")[0]
    row_string += "'"+relat_name+"', '"+format(time_taken, '.2f')+"', '" 
    print("k: ", k)
    print("McAiger computation time: ", time_taken)
    

    return k,time_taken
############ main #############

modelFile = open (sys.argv[1], "rb")

use_kissat = 1
# if(len(sys.argv) > 2):
#     if sys.argv[2] == "-kis":
#         use_kissat = 1
only_gen_kwm = 0
# if(len(sys.argv) > 3):
#     if sys.argv[2] == "-kwm":
#         print("HI")
#         only_gen_kwm = 1
#         k = str(sys.argv[3])

# k,mc_time = mcaiger(k, modelFile)
k = str(sys.argv[2])
mc_time = 0.0

if (k=="1"):
    kwFile = sys.argv[1]
    t2,con1, con2, con3, s0,s1,s2 = generate_proof(kwFile)
    print(" ")
    print("######################################################")

    row_string += str(t2) + "', '"
    print(" ")
    print("######################################################")

    t,cnf1 = aiger2cnf(con1)
    print("result1: ")
    row_string+=str(t)+"', '" + s0 +"'"
    res1 = sat(cnf1, 0)
    print(" ")
    print("######################################################")

    t, cnf2 = aiger2cnf(con2)
    print("result2: ")
    row_string+=", '"+ str(t) +"', '"+s1+"'"

    res2 = sat(cnf2, 1)
    print(" ")
    print("######################################################")

    t,cnf3 = aiger2cnf(con3)
    print("result3: ")
    row_string+=", '"+ str(t) +"', '"+s2+"'"

    res3 = sat(cnf3, 2)
    print("######################################################")
    if(res1 == "s UNSATISFIABLE"and res2 == "s UNSATISFIABLE" and res3 == "s UNSATISFIABLE"):
        print("SUCCESS")
        print(row_string)
    else:
        print("Certification check failed.")
    quit(0)



if(len(sys.argv) > 3):
    if sys.argv[3] == "-nmc":
        k = sys.argv[4]

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(name)-12s %(message)s', datefmt='%m-%d %H:%M', filename=sys.path[0] + '/temp/log_'+k+'.txt', filemode='w')
console = logging.StreamHandler()
console.setLevel(logging.INFO)
# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger('').addHandler(console)
logging.info(modelFile.name)
logging.info("k="+k)

print("reading model")
model_file = open(sys.path[0] + "/" + modelFile.name ,"r",encoding="utf8", errors='ignore')
h = model_file.readline()
print("header:", h)
header = h.split(" ")
fsize_ands = header[-1]
fsize_maxvar = format(int(header[1])/1000, '.2f')
model_file.close()

row_string += fsize_maxvar + "', '"+k+"', '"
logger = logging.getLogger("mcaiger")
logger.info("time="+str(mc_time))
logger.info("maxvar="+fsize_maxvar)
print("###########################\n Certifying...\n###########################")
kwFile = generate_kw_model(k, modelFile)
print(" ")
print("######################################################")

if only_gen_kwm ==1:
    print(row_string)
if only_gen_kwm ==0:
    print("checking combinational simulation....")
    t1,out1, out2, out3, s3,s4,s5 = check_comb_sim(k, modelFile, kwFile)
    row_string += str(t1) + "', '"
    print(" ")

    t2,con1, con2, con3, s0,s1,s2 = generate_proof(kwFile)
    print(" ")
    print("######################################################")

    row_string += str(t2) + "', '"
    print(" ")
    print("######################################################")

    t,cnf1 = aiger2cnf(con1)
    print("result1: ")
    row_string+=str(t)+"', '" + s0 +"'"
    res1 = sat(cnf1, 0)
    print(" ")
    print("######################################################")

    t, cnf2 = aiger2cnf(con2)
    print("result2: ")
    row_string+=", '"+ str(t) +"', '"+s1+"'"

    res2 = sat(cnf2, 1)
    print(" ")
    print("######################################################")

    t,cnf3 = aiger2cnf(con3)
    print("result3: ")
    row_string+=", '"+ str(t) +"', '"+s2+"'"

    res3 = sat(cnf3, 2)
    print("######################################################")

    t,c1 = aiger2cnf(out1)
    print("combination simulation F==F': ")
    row_string+=", '" + str(t) +"', '"+s3+"'"

    res4 = sat(c1, 3)
    print(" ")
    print("######################################################")


    print("combination simulation p'=> p: ")
    t, c2 = aiger2cnf(out2)
    row_string+=", '" + str(t)+ "', '"+s4+"'"

    res5 = sat(c2, 4)
    print(" ")
    print("######################################################")

    print("combination simulation R=>R': ")
    #res6 = verify_qbf(out3, 5)
    t, c3 = aiger2cnf(out3)
    row_string+=", '" + str(t)+ "', '"+s4+"'"

    res6 = sat(c3, 4)
    print(" ")
    print("######################################################")


    if(res1 == "s UNSATISFIABLE"and res2 == "s UNSATISFIABLE" and res3 == "s UNSATISFIABLE"and res4 == "s UNSATISFIABLE" and res5 == "s UNSATISFIABLE" and res6=="s UNSATISFIABLE" ):
        print("SUCCESS")
        print(row_string)
    else:
        print("Certification check failed.")

